/*
    Bantumi
    Copyright 2005 - 2007 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#ifndef __SDPRECORD_H
#define __SDPRECORD_H

#include <stack>
#include <vector>


class SdpRecord {
public:
	SdpRecord();
	~SdpRecord();

	SdpRecord* addNil();

	SdpRecord* addUint8(unsigned char);
	SdpRecord* addUint16(unsigned short);
	SdpRecord* addUint32(unsigned int);
	SdpRecord* addUint64(unsigned long long);
	SdpRecord* addUint(unsigned long long);

	SdpRecord* addSint8(signed char);
	SdpRecord* addSint16(signed short);
	SdpRecord* addSint32(signed int);
	SdpRecord* addSint64(signed long long);
	SdpRecord* addSint(signed long long);

	SdpRecord* addUuid16(unsigned short);
	SdpRecord* addUuid32(unsigned int);
	SdpRecord* addUuid128(const unsigned char*);

	SdpRecord* addString(const char*);

	SdpRecord* addBoolean(bool);

	SdpRecord* addDES();
	SdpRecord* endDES();

	SdpRecord* addDEA();
	SdpRecord* endDEA();

	SdpRecord* addUrl(const char*);

	SdpRecord* startRecord();
	SdpRecord* endRecord();

	SdpRecord* addAttribute(unsigned int id);
	SdpRecord* endAttribute();

	const unsigned char* getData();
	int getLength();

private:
	class Context {
	public:
		Context(int l);
		Context(const Context& c);
		Context& operator=(const Context& c);
		~Context();
		void assureSpace(int s);
		void writeByte(unsigned char b);
		void writeBytes(const unsigned char* bytes, int len);
		const unsigned char* getData();
		int getLength();

	private:
		unsigned char* array;
		int pos;
		int len;
	};

	std::stack<Context> contexts;
	Context& top();
	void writeType(int type, int size);
	void writeUint8(unsigned char value);
	void writeUint16(unsigned short value);
	void writeUint32(unsigned int value);
	void writeData(int type, const unsigned char* data, int len);
};


class SdpRecordParser {
public:
	class Record;
	class Buffer;
	class ArrayRecord;
	class UuidRecord;

	SdpRecordParser(const unsigned char* data, int len);
	~SdpRecordParser();
	Record* parseRecord(Buffer& buffer);
	unsigned char readUint8(Buffer& buffer);
	unsigned short readUint16(Buffer& buffer);
	unsigned int readUint32(Buffer& buffer);
	unsigned long long readUint64(Buffer& buffer);
	Record* findAttribute(unsigned int attribute) const;
	bool hasServiceUuid16(unsigned int uuid) const;
	bool hasServiceUuid32(unsigned int uuid) const;
	bool hasServiceUuid128(const unsigned char* uuid) const;
	bool hasService(const UuidRecord& uuid) const;

	class AttributeVisitor {
	public:
		virtual ~AttributeVisitor() {}
		virtual void arrayStart(const ArrayRecord* r) {}
		virtual void arrayEnd(const ArrayRecord* r) {}
		virtual void visitRecord(const Record* r) {}
		virtual void done() {}
	};

	enum Verb { CHECK, READ, FINISH };
	enum RecordType { DES, DEA, UUID, PRIMITIVE, STRING, NIL, END };

	class SerializeVisitor : public AttributeVisitor {
	public:
		void arrayStart(const ArrayRecord* r);
		void arrayEnd(const ArrayRecord* r);
		void visitRecord(const Record* r);

		std::vector<RecordType> events;
		std::vector<const Record*> records;
	};

	struct ScriptBlock {
		Verb verb;
		RecordType type;
		Record* record;
	};
	bool checkScript(const ScriptBlock* block, const Record* record);
	bool getRfcommChannel(int* channel);
	void visitRecords(AttributeVisitor* visitor, const Record* record);
	void visitAttribute(AttributeVisitor* visitor, const Record* attribute);


public:
	class Record {
	public:
		int getType() const;
		int getSize() const;
	protected:
		Record(int type, int size = 0);
		int type;
		int size;
	};
	class NilRecord : public Record {
	public:
		NilRecord();
	};
	class ArrayRecord : public Record {
	public:
		ArrayRecord(int type);
		~ArrayRecord();
		void appendRecord(Record* r);
		int getCount() const;
		Record* getRecord(int index) const;
	private:
		std::vector<Record*> children;
	};
	class StringRecord : public Record {
	public:
		StringRecord();
		StringRecord(int type, Buffer& buffer);
		~StringRecord();
		const char* getString() const;
		void set(int type, const char* str);
	private:
		char* string;
	};
	class PrimitiveRecord : public Record {
	public:
		PrimitiveRecord();
		PrimitiveRecord(int type, int size, unsigned long long value);
		unsigned long long getValue() const;
	private:
		unsigned long long value;
	};
	class UuidRecord : public Record {
	public:
		UuidRecord();
		UuidRecord(int size, unsigned int uuid);
		UuidRecord(int size, Buffer& buffer);
		bool matches(const UuidRecord& uuid) const;
	private:
		unsigned int uuid;
		unsigned char uuid128[16];
	};

	class Buffer {
	public:
		Buffer(const unsigned char* data, int len);
		int getLength() const;
		const unsigned char* getData() const;
		void consume(int bytes);
		unsigned char readByte();
		Buffer readSubstring(int bytes);
	private:
		const unsigned char* data;
		int length;
	};
private:

	void raiseError();

	Record* rootRecord;
	bool error;
};

#endif
