/*
    Bantumi
    Copyright 2005 - 2007 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#include "glfont.h"
#include <ft2build.h>
#include FT_FREETYPE_H

#define FLOOR64(x) ((x) / 64)
#define CEIL64(x) FLOOR64(x + 63)

GLFont *makeGLFont(const char *name, int size, int w, int h, unsigned char first, unsigned char last, unsigned char skipstart, unsigned char skiplast) {
	int offset = first;
	int n = last - offset + 1;
#ifndef EXPORT_FONT
	TexGLFont *f = new TexGLFont(offset, n);
#else
	char* f = NULL;
	char namebuf[100];
	snprintf(namebuf, sizeof(namebuf), "%s-%d-%dx%d.txt", name, size, w, h);
	FILE* info = fopen(namebuf, "w");
	fprintf(info, "%d %d\n", offset, n);
#endif

	FT_Library library;
	FT_Error error = FT_Init_FreeType(&library);
	if (error) {
		fprintf(stderr, "FreeType error: %d\n", error);
		delete f;
		return NULL;
	}

	FT_Face face;
	error = FT_New_Face(library, name, 0, &face);
	if (error) {
		fprintf(stderr, "FreeType error: %d\n", error);
		FT_Done_FreeType(library);
		delete f;
		return NULL;
	}

	if (!FT_IS_SCALABLE(face)) {
		fprintf(stderr, "Non-scalable fonts not supported\n");
		FT_Done_Face(face);
		FT_Done_FreeType(library);
		delete f;
		return NULL;
	}
	error = FT_Set_Char_Size(face, 0, size*64, 0, 0);
	if (error) {
		fprintf(stderr, "FreeType error: %d\n", error);
		FT_Done_Face(face);
		FT_Done_FreeType(library);
		delete f;
		return NULL;
	}


	unsigned char* buffer = new unsigned char[w*h];
	int bufferWidth = w;
	int bufferHeight = h;
	memset(buffer, 0, bufferWidth*bufferHeight);


	int margin = 1;

	int cury = 0;
	int curx = 0;
	int nexty = cury;
	for (int c = offset; c < offset+n; c++) {
		if (c >= skipstart && c <= skiplast)
			continue;
		int minx, maxx, miny, maxy, adv;

		FT_UInt index = FT_Get_Char_Index(face, c);
		error = FT_Load_Glyph(face, index, FT_LOAD_DEFAULT);
		if (error)
			break;

		FT_Glyph_Metrics* metrics = &face->glyph->metrics;

		minx = FLOOR64(metrics->horiBearingX);
		maxx = minx + CEIL64(metrics->width);
		maxy = FLOOR64(metrics->horiBearingY);
		miny = maxy - CEIL64(metrics->height);
		adv = CEIL64(metrics->horiAdvance);

		error = FT_Render_Glyph(face->glyph, FT_RENDER_MODE_NORMAL);
		if (error)
			break;

		int width = face->glyph->bitmap.width;
		int height = face->glyph->bitmap.rows;

		if ((curx + width + margin > bufferWidth && nexty + height + margin > bufferHeight) || cury + height + margin > bufferHeight)
			break;
		if (curx + width + margin > bufferWidth) {
			curx = 0;
			cury = nexty;
		}
		for (int y = 0; y < height; y++) {
			memcpy(buffer + (cury + y)*bufferWidth + curx, face->glyph->bitmap.buffer + face->glyph->bitmap.pitch*y, width);
		}
#ifndef EXPORT_FONT
		f->setParams(c, curx, cury, width, height, adv, minx, miny);
#else
		fprintf(info, "%d %d %d %d %d %d %d %d\n", c, curx, cury, width, height, adv, minx, miny);
#endif

		if (cury + height + margin > nexty)
			nexty = cury + height + margin;
		curx += width + margin;
	}
	if (error) {
		fprintf(stderr, "FreeType error: %d\n", error);
		FT_Done_Face(face);
		FT_Done_FreeType(library);
		delete [] buffer;
		return NULL;
	}

#ifndef EXPORT_FONT
	f->setAlphaImage(buffer, bufferWidth, bufferHeight);
#else
	snprintf(namebuf, sizeof(namebuf), "%s-%d-%dx%d.pgm", name, size, w, h);
	FILE* bitmap = fopen(namebuf, "wb");
	fprintf(bitmap, "P5\n%d %d\n%d\n", w, h, 255);
	fwrite(buffer, 1, bufferWidth*bufferHeight, bitmap);
	fclose(bitmap);
	fclose(info);
#endif
	delete [] buffer;

	FT_Done_Face(face);
	FT_Done_FreeType(library);

#ifndef EXPORT_FONT
	return f;
#else
	return NULL;
#endif
}

#ifdef EXPORT_FONT
int main(int argc, char* argv[]) {
	makeGLFont("Vera.ttf", 26, 64, 64, '0', '9');
	makeGLFont("Vera.ttf", 30, 256, 256, ' ', '\366', 'z' + 1, '\366' - 1);
	return 0;
}
#endif

