/*
    Bantumi
    Copyright 2005 - 2007 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#ifdef SDL_EGL
#include <SDL_syswm.h>
#include <EGL/egl.h>
#else
#include <SDL_opengl.h>
#endif
#include <math.h>
#include <string.h>
#include <time.h>
#include "bantumigl.h"
#include "bantumi.h"
#include "glfont.h"
#ifdef __APPLE__
#include <CoreFoundation/CoreFoundation.h>
#endif
#include "tcpip.h"
#include "gui.h"

#ifndef FONTNAME
#define FONTNAME "Vera.ttf"
#endif

int width = 640;
int height = 480;

void makeScreenShot();

static BantumiGL *fe;
static int lasttime;

#ifdef SDL_EGL
EGLDisplay display;
EGLSurface surface;
EGLContext context;

const EGLint attribList[] = {
	EGL_DEPTH_SIZE, 16,
	EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
	EGL_NONE
};

bool initEGL() {
	SDL_SysWMinfo wminfo;
	SDL_VERSION(&wminfo.version)
	SDL_GetWMInfo(&wminfo);
	SDL_ShowCursor(0);

	display = eglGetDisplay((EGLNativeDisplayType) wminfo.info.x11.display);
	if (display == EGL_NO_DISPLAY) {
		fprintf(stderr, "eglGetDisplay failed, %d\n", eglGetError());
		return false;
	}
	if (eglInitialize(display, NULL, NULL) == EGL_FALSE) {
		fprintf(stderr, "eglInitialize failed, %d\n", eglGetError());
		return false;
	}
	eglBindAPI(EGL_OPENGL_ES_API);

	EGLConfig config;
	EGLint numConfigs;
	if (eglChooseConfig(display, attribList, &config, 1, &numConfigs) == EGL_FALSE) {
		fprintf(stderr, "eglChooseConfig failed, %d\n", eglGetError());
		return false;
	}

	context = eglCreateContext(display, config, NULL, NULL);
	if (!context) {
		fprintf(stderr, "eglCreateContext failed, %d\n", eglGetError());
		return false;
	}
	surface = eglCreateWindowSurface(display, config, (void*) wminfo.info.x11.window, NULL);
	if (!surface) {
		fprintf(stderr, "eglCreateWindowSurface failed, %d\n", eglGetError());
		return false;
	}

	if (eglMakeCurrent(display, surface, surface, context) == EGL_FALSE) {
		fprintf(stderr, "eglMakeCurrent failed, %d\n", eglGetError());
		return false;
	}

	return true;
}

void closeEGL() {
	eglMakeCurrent(display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
	eglDestroySurface(display, surface);
	eglDestroyContext(display, context);
	eglTerminate(display);
	eglReleaseThread();
}

void swapBuffers() {
	eglSwapBuffers(display, surface);
}
#else
void swapBuffers() {
	SDL_GL_SwapBuffers();
}
#endif


void redraw() {
	fe->draw(lasttime);
	swapBuffers();
}

int main(int argc, char *argv[]) {
	if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE)<0) {
		fprintf(stderr,"%s\n",SDL_GetError());
		return 1;
	}
	if (atexit(SDL_Quit)) {
		fprintf(stderr,"atexit(...)\n");
		SDL_Quit();
		return 1;
	}
//	width = 176;
//	height = 208;
#ifdef SDL_EGL
	width = height = 0;
	int flags = SDL_FULLSCREEN;
#else
#ifdef SCREENSHOT
	width = 1024;
	height = 1280;
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 1);
#endif
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	bool fullscreen = false;
	int flags = SDL_OPENGL | SDL_RESIZABLE | (fullscreen ? SDL_FULLSCREEN : 0);
#endif
	SDL_WM_SetCaption("Bantumi GL", "Bantumi GL");
	SDL_Surface *screen = SDL_SetVideoMode(width, height, 0, flags);
	if (screen == NULL) {
		fprintf(stderr,"%s\n",SDL_GetError());
		return 1;
	}

#ifdef SDL_EGL
	if (!initEGL())
		return 1;
#endif

	srand(time(NULL));
	initSockets();
	initGui();

	fe = new BantumiGL(screen->w, screen->h);
	GLFont *textFont, *numFont;
	const char *fontname = FONTNAME;
#ifdef __APPLE__
	char fontPath[250];
	CFURLRef url = CFBundleCopyResourcesDirectoryURL(CFBundleGetMainBundle());
	CFURLGetFileSystemRepresentation(url, true, (UInt8*) fontPath, sizeof(fontPath));
	CFRelease(url);
	strcat(fontPath, "/");
	strcat(fontPath, fontname);
	fontname = fontPath;

	CFPropertyListRef levelRef = CFPreferencesCopyAppValue(CFSTR("gameLevel"), kCFPreferencesCurrentApplication);
	CFPropertyListRef numRef = CFPreferencesCopyAppValue(CFSTR("gameNum"), kCFPreferencesCurrentApplication);
	if (levelRef && CFGetTypeID(levelRef) == CFNumberGetTypeID()) {
		int value;
		if (CFNumberGetValue((CFNumberRef) levelRef, kCFNumberIntType, &value))
			fe->setLevel(value);
	}
	if (numRef && CFGetTypeID(numRef) == CFNumberGetTypeID()) {
		int value;
		if (CFNumberGetValue((CFNumberRef) numRef, kCFNumberIntType, &value))
			fe->setNum(value);
	}
	if (levelRef)
		CFRelease(levelRef);
	if (numRef)
		CFRelease(numRef);
#endif
	
#define makeSetFonts() \
	numFont = makeGLFont(fontname, 26, 64, 64, '0', '9');		\
	textFont = makeGLFont(fontname, 30, 256, 256, ' ', '\366', 'z'+1, '\366'-1);	\
	fe->setTextFont(textFont);					\
	fe->setNumFont(numFont);
	makeSetFonts();
	Bantumi *b = new Bantumi(fe);

#ifdef SCREENSHOT
	fe->draw(0);
	fe->draw(10000);
	fe->draw(20000);
	fe->draw(30000);
	makeScreenShot();
#endif

	bool active = true;
	int frames = 0;
	int starttime = SDL_GetTicks();
	lasttime = starttime;
	while (true) {
		SDL_Event event;
		bool done = false;
		while (SDL_PollEvent(&event)) {
			if (event.type == SDL_VIDEORESIZE) {
				screen = SDL_SetVideoMode(event.resize.w, event.resize.h, screen->format->BitsPerPixel, flags);
				if (screen == NULL) {
					fprintf(stderr, "%s\n", SDL_GetError());
					return 1;
				}
				delete numFont;
				delete textFont;
				makeSetFonts();
				fe->reinitGL(screen->w, screen->h);
			} else if (event.type == SDL_KEYDOWN) {
				switch (event.key.keysym.sym) {
				case SDLK_ESCAPE:
				case SDLK_BACKSPACE:
				case SDLK_q:
					b->pressed(ESCAPE);
					break;
/*				case SDLK_F1:
					fullscreen = !fullscreen;
					if (fullscreen) flags |= SDL_FULLSCREEN;
					else flags &= ~SDL_FULLSCREEN;
					screen = SDL_SetVideoMode(screen->w, screen->h, screen->format->BitsPerPixel, flags);
					if (screen == NULL) {
						fprintf(stderr,"%s\n",SDL_GetError());
						return 1;
					}
					delete numFont;
					delete textFont;
					makeSetFonts();
					fe->reinitGL(screen->w, screen->h);
					break;
*/				case SDLK_LEFT:
					b->pressed(LEFT);
					break;
				case SDLK_RIGHT:
					b->pressed(RIGHT);
					break;
				case SDLK_UP:
					b->pressed(UP);
					break;
				case SDLK_DOWN:
					b->pressed(DOWN);
					break;
				case SDLK_SPACE:
				case SDLK_RETURN:
					b->pressed(SELECT);
					break;
				default:
					break;
				}
			} else if (event.type == SDL_QUIT) {
				done = true;
			} else if (event.type == SDL_ACTIVEEVENT) {
				if (event.active.state == SDL_APPACTIVE)
					active = event.active.gain;
			} else if (event.type == SDL_MOUSEBUTTONDOWN) {
				if (event.button.button == SDL_BUTTON_RIGHT)
					b->pressed(ESCAPE);
				else
					b->click(event.button.x, event.button.y);
			} else if (event.type == SDL_MOUSEBUTTONUP) {
			} else if (event.type == SDL_MOUSEMOTION) {
			}
		}
		if (done) break;
		if (b->exit()) break;


		b->update();

		int time = SDL_GetTicks() - starttime;
		int difftime = time - lasttime;
		lasttime = time;
		if (difftime < 15)
			SDL_Delay(10);

		fe->draw(time);
		swapBuffers();
		frames++;
		if (active) {
			SDL_Delay(1);
		} else SDL_Delay(50);
//		SDL_Delay(500);

	}
//	printf("%d fps\n", 1000*frames/(SDL_GetTicks() - starttime));

#ifdef __APPLE__
	int level = fe->getLevel();
	int num = fe->getNum();
	levelRef = CFNumberCreate(NULL, kCFNumberIntType, &level);
	numRef = CFNumberCreate(NULL, kCFNumberIntType, &num);
	CFPreferencesSetAppValue(CFSTR("gameLevel"), levelRef, kCFPreferencesCurrentApplication);
	CFPreferencesSetAppValue(CFSTR("gameNum"), numRef, kCFPreferencesCurrentApplication);
	CFRelease(levelRef);
	CFRelease(numRef);
	CFPreferencesAppSynchronize(kCFPreferencesCurrentApplication);
#endif

	delete b;
	delete fe;
	delete textFont;
	delete numFont;

	closeSockets();

#ifdef SDL_EGL
	closeEGL();
#endif

	return 0;
}


